import torch
from transformers import CLIPProcessor, CLIPModel
from amber_mcap.automation.annotation import ImageAnnotation
from typing import Tuple, List, Optional
from amber_mcap.util.lvis.lvis_v1_categories import (
    LVIS_CATEGORIES as LVIS_V1_CATEGORIES,
)
from PIL.Image import Image


class ClipEncoder:
    def __init__(self, model: str = "ViT-B/32"):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        # self.model, self.preprocess = load("ViT-B/32", device=self.device)
        self.model = CLIPModel.from_pretrained("openai/clip-vit-base-patch32")
        self.processor = CLIPProcessor.from_pretrained("openai/clip-vit-base-patch32")
        self.lvis_classes: List[str] = [k["synonyms"][0] for k in LVIS_V1_CATEGORIES]
        self.lvis_prompts: List[str] = []
        for lvis_class in self.lvis_classes:
            self.lvis_prompts.append("A photo of a " + lvis_class)
        self.lvis_text_embeddings: Optional[torch.Tensor] = None

    def get_single_image_embeddings(self, image: Image) -> torch.Tensor:
        with torch.no_grad():
            inputs = self.processor(
                text=None, images=[image], return_tensors="pt", padding=True
            )
            return self.model.get_image_features(pixel_values=inputs["pixel_values"])

    def get_single_image_embeddings_for_objects(
        self, image: Image, annotation: ImageAnnotation
    ) -> ImageAnnotation:
        with torch.no_grad():
            for bounding_box in annotation.bounding_boxes:
                bounding_box.clip_embeddings = self.get_single_image_embeddings(
                    image.crop(
                        (
                            int(bounding_box.box.x1),
                            int(bounding_box.box.y1),
                            int(bounding_box.box.x2),
                            int(bounding_box.box.y2),
                        )
                    )
                ).tolist()[0]
                assert len(bounding_box.clip_embeddings) == 512
        return annotation

    def get_single_text_embeddings(self, text: str) -> torch.Tensor:
        with torch.no_grad():
            inputs = self.processor(
                text=[text], images=None, return_tensors="pt", padding=True
            )
            return self.model.get_text_features(
                inputs["input_ids"], inputs["attention_mask"]
            )

    def get_text_embeddings(self, texts: List[str]) -> torch.Tensor:
        with torch.no_grad():
            inputs = self.processor(
                text=texts, images=None, return_tensors="pt", padding=True
            )
            return self.model.get_text_features(
                inputs["input_ids"], inputs["attention_mask"]
            )

    def classify_with_custom_vocabulary(
        self, image_embeddings: torch.Tensor, texts: List[str]
    ) -> Optional[Tuple[str, float]]:
        if self.lvis_text_embeddings == None:
            with torch.no_grad():
                inputs = self.processor(
                    text=[self.lvis_prompts],
                    images=None,
                    return_tensors="pt",
                    padding=True,
                )
                self.lvis_text_embeddings = self.model.get_text_features(
                    inputs["input_ids"], inputs["attention_mask"]
                )
        prompts: List[str] = []
        for text in texts:
            prompts.append("A photo of a " + text)
        with torch.no_grad():
            text_embeddings = torch.cat(
                [self.lvis_text_embeddings, self.get_text_embeddings(prompts)],
                dim=0,
            )
            image_embeddings /= image_embeddings.norm(dim=-1, keepdim=True)
            text_embeddings /= text_embeddings.norm(dim=-1, keepdim=True)
            similarity = (
                image_embeddings.to(torch.float32) @ text_embeddings.to(torch.float32).T
            ).softmax(dim=-1)
            values, indices = similarity.topk(1)
            for value, index in zip(values, indices):
                if index < len(self.lvis_classes):
                    return None
                else:
                    return (texts[index - len(self.lvis_classes)], value.item())
        return None

    def get_single_text_embeddings_for_positive_negative_prompts(
        self, object_name: str
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        return (
            self.get_single_text_embeddings("A photo of a " + object_name),
            self.get_single_text_embeddings("Not a photo of a " + object_name),
        )


if __name__ == "__main__":
    pass



import os
from torch.utils.data import IterableDataset
from typing import Any, List, Optional
from yaml import safe_load  # type: ignore
from amber_mcap.exception import CertificationError
from dataclasses import dataclass
from dataclass_wizard import JSONWizard
import glob
import boto3
import requests  # type: ignore
import datetime


@dataclass
class MessageMetaData(JSONWizard):  # type: ignore
    topic: str = ""
    rosbag_path: str = ""
    publish_time: datetime.datetime = datetime.datetime.fromtimestamp(
        0, datetime.timezone.utc
    )  # Unix epoch time


class Rosbag2Dataset(IterableDataset):  # type: ignore
    message_metadata: List[MessageMetaData] = []
    compressed: bool = False

    def __init__(
        self,
        rosbag_path: str,
        compressed: bool,
        transform: Any = None,
        target_transform: Any = None,
    ) -> None:
        if os.path.isfile(rosbag_path):
            self.rosbag_files = [rosbag_path]
        else:
            self.rosbag_files = glob.glob(rosbag_path + "/**/*.mcap", recursive=True)
        self.compressed = compressed
        self.transform = transform
        self.target_transform = target_transform
        self.message_metadata.clear()

    def get_sample_data_index_by_timestamp(
        self,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        data_duration: datetime.timedelta,
    ) -> List[int]:
        last_publish_timestamp_in_sample: datetime.datetime = (
            datetime.datetime.fromtimestamp(0, datetime.timezone.utc)
        )  # Unix epoch time
        indices: List[int] = []
        assert len(self.message_metadata) == len(self)
        for i in range(len(self.message_metadata)):
            if (
                self.get_metadata(i).publish_time > start_time
                and self.get_metadata(i).publish_time < end_time
            ):
                if len(indices) == 0:
                    indices.append(i)
                    last_publish_timestamp_in_sample = self.get_metadata(i).publish_time
                elif (
                    last_publish_timestamp_in_sample + data_duration
                    <= self.get_metadata(i).publish_time
                ):
                    indices.append(i)
                    last_publish_timestamp_in_sample = self.get_metadata(i).publish_time
        return indices

    def get_metadata(self, index: int) -> MessageMetaData:
        return self.message_metadata[index]

    def get_first_timestamp(self) -> datetime.datetime:
        assert len(self.message_metadata) != 0
        return min(self.message_metadata, key=lambda x: x.publish_time).publish_time

    def get_last_timestamp(self) -> datetime.datetime:
        assert len(self.message_metadata) != 0
        return max(self.message_metadata, key=lambda x: x.publish_time).publish_time


def download_rosbag(
    bucket_name: str,
    remote_rosbag_directory: str,
    remote_rosbag_filename: str,
    endpoint_url: str,
    download_dir: str = "/tmp/amber/remote_bags",
    is_public: bool = True,
    false_overwrite: bool = False,
    aws_access_key_id: Optional[str] = os.getenv("AWS_ACCESS_KEY_ID"),
    aws_secret_access_key: Optional[str] = os.getenv("AWS_SECRET_ACCESS_KEY"),
) -> str:
    if aws_access_key_id is None:
        raise CertificationError("You should specify aws access key.")
    if aws_secret_access_key is None:
        raise CertificationError("You should specify aws secret access key.")
    if is_public:
        s3 = boto3.client(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        remote_rosbag_path = os.path.join(
            remote_rosbag_directory, remote_rosbag_filename
        )
        local_rosbag_path = os.path.join(download_dir, remote_rosbag_path)
        if os.path.exists(local_rosbag_path) and not false_overwrite:
            return local_rosbag_path
        with open(local_rosbag_path, mode="wb") as f:
            f.write(
                requests.get(
                    s3.generate_presigned_url(
                        ClientMethod="get_object",
                        Params={"Bucket": bucket_name, "Key": remote_rosbag_path},
                        ExpiresIn=3600,
                        HttpMethod="GET",
                    )
                ).content
            )
        return local_rosbag_path
    else:
        s3 = boto3.resource(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        remote_rosbag_path = os.path.join(
            remote_rosbag_directory, remote_rosbag_filename
        )
        bucket = s3.Bucket(bucket_name)
        local_rosbag_path = os.path.join(download_dir, remote_rosbag_path)
        if os.path.exists(local_rosbag_path) and not false_overwrite:
            return local_rosbag_path
        bucket.download_file(remote_rosbag_path, local_rosbag_path)
        return local_rosbag_path



import shutil
from distutils import log as distutils_log
from pathlib import Path
from typing import Any, Dict

import skbuild
import skbuild.constants
import subprocess
import sysconfig
import os

__all__ = ("build",)


def get_poetry_venv_path() -> Path:
    """Get the path to the Poetry virtual environment."""
    try:
        result = subprocess.run(
            ["poetry", "env", "info", "--path"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
        venv_path = result.stdout.strip()
        return Path(venv_path)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Failed to get Poetry virtual environment: {e.stderr}"
        ) from e


def build(setup_kwargs: Dict[str, Any]) -> None:
    venv_path = get_poetry_venv_path()
    """Build C-extensions."""
    pybind11_path = (
        venv_path
        / "lib"
        / f"python{sysconfig.get_python_version()}"
        / "site-packages"
        / "pybind11"
        / "share"
        / "cmake"
        / "pybind11"
    )
    skbuild.setup(
        **setup_kwargs,
        script_args=["build_ext"],
        cmake_args=["-Dpybind11_DIR=" + str(pybind11_path)],
    )

    src_dir = Path(skbuild.constants.CMAKE_INSTALL_DIR())

    # Get the installation directory for the Poetry virtual environment
    dest_dir = (
        venv_path / "lib" / f"python{sysconfig.get_python_version()}" / "site-packages"
    )

    print("Copy from " + str(src_dir) + " to " + str(dest_dir))

    # Delete C-extensions copied in previous runs, just in case.
    # remove_files(dest_dir, "**/*.pyd")
    remove_files(dest_dir, "**/tf2_amber.*.so")

    # Copy built C-extensions back to the Poetry virtual environment.
    copy_files(src_dir, dest_dir, "**/*.pyd")
    copy_files(src_dir, dest_dir, "**/*.so")


def remove_files(target_dir: Path, pattern: str) -> None:
    """Delete files matched with a glob pattern in a directory tree."""
    for path in target_dir.glob(pattern):
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()
        distutils_log.info(f"removed {path}")


def copy_files(src_dir: Path, dest_dir: Path, pattern: str) -> None:
    """Copy files matched with a glob pattern in a directory tree to another."""
    for src in src_dir.glob(pattern):
        dest = dest_dir / src.relative_to(src_dir)
        if src.is_dir():
            # NOTE: inefficient if subdirectories also match to the pattern.
            copy_files(src, dest, "*")
        else:
            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(src, dest)
            distutils_log.info(f"copied {src} to {dest}")


if __name__ == "__main__":
    build({})



from enum import Enum


class TimeUnit(Enum):
    NANOSECOND = 1e-09
    """Type of nanosecond, the corresponding value indicates the number of seconds per unit.
    """
    MICROSECOND = 1e-06
    """Type of microsecond, the corresponding value indicates the number of seconds per unit.
    """
    MILLISECOND = 1e-3
    """Type of millisecond, the corresponding value indicates the number of seconds per unit.
    """
    SECOND = 1
    """Type of second, the corresponding value indicates the number of seconds per unit.
    """


class Time:
    __value: float = 0.0
    __unit: TimeUnit = TimeUnit.SECOND

    def __init__(self, value: float, unit: TimeUnit) -> None:
        self.value = value * unit.value

    def get(self, unit: TimeUnit) -> float:
        return float(self.value) / float(unit.value)



from amber_mcap.dataset.images_and_annotations_dataset import (
    ImagesAndAnnotationsDataset,
)
from amber_mcap.automation.automation import Automation
from amber_mcap.automation.annotation import ImageAnnotation, BoundingBoxAnnotation
from amber_mcap.automation.clip_encoder import ClipEncoder
from amber_mcap.automation.sentence_transformer import TextEncoder
from amber_mcap.automation.task_description import (
    ClipImageAnnotationFilterConfig,
    ClipClassifyMethod,
)
from torchvision import transforms
from typing import List, Dict, Tuple
from torch.nn.functional import cosine_similarity
import torch
from amber_mcap.exception import RuntimeError
import copy


class ClipImageAnnotationFilter(Automation):  # type: ignore
    def __init__(self, yaml_path: str) -> None:
        self.config = ClipImageAnnotationFilterConfig.from_yaml_file(yaml_path)
        self.clip_encoder = ClipEncoder()
        self.text_encoder = TextEncoder()
        self.text_embeddings: Dict[str, Tuple[torch.Tensor, torch.Tensor]] = {}
        for target_object in self.config.target_objects:
            self.text_embeddings[
                target_object
            ] = self.clip_encoder.get_single_text_embeddings_for_positive_negative_prompts(
                target_object
            )
        self.to_pil_image = transforms.ToPILImage()
        return

    def inference_with_lvis_vocabulary(
        self,
        bounding_box: BoundingBoxAnnotation,
    ) -> bool:
        clip_embeddings: torch.Tensor = torch.tensor(
            bounding_box.clip_embeddings, dtype=float
        )
        result = self.clip_encoder.classify_with_custom_vocabulary(
            clip_embeddings, self.config.target_objects
        )
        if result != None:
            return True
        else:
            return False

    def inference_with_bert(
        self,
        bounding_box: BoundingBoxAnnotation,
    ) -> bool:
        annotation_text_embeddings = self.clip_encoder.get_single_text_embeddings(
            "A photo of a " + bounding_box.object_class
        )
        clip_embeddings: torch.Tensor = torch.tensor(
            bounding_box.clip_embeddings, dtype=float
        )
        for target_object in self.config.target_objects:
            clip_similarity = cosine_similarity(
                clip_embeddings / torch.sum(clip_embeddings),
                self.text_embeddings[target_object][0],
            )
            positive = cosine_similarity(
                clip_embeddings / torch.sum(clip_embeddings)
                + annotation_text_embeddings
                / torch.sum(annotation_text_embeddings)
                * self.text_encoder.cosine_similarity(
                    bounding_box.object_class, target_object
                ),
                self.text_embeddings[target_object][0],
            )
            negative = cosine_similarity(
                clip_embeddings / torch.sum(clip_embeddings),
                # - annotation_text_embeddings
                # / torch.sum(annotation_text_embeddings)
                # * bounding_box.score
                # * self.text_encoder.cosine_similarity(
                #     bounding_box.object_class, target_object
                # ),
                self.text_embeddings[target_object][1],
            )
            # print("P/N ratio : " + str(positive.item() / negative.item()))
            # print("Score : " + str(bounding_box.score))
            # print("Class : " + str(bounding_box.object_class))
            # print("Clip Cosine Similarity with bert: " + str(positive.item()))
            # print("Clip Cosine Similarity without bert: " + str(clip_similarity.item()))
            # print("")
            if (
                positive.item()
                > self.config.consider_annotation_with_bert_config.positive_nagative_ratio
                * negative.item()
                and positive.item()
                >= self.config.consider_annotation_with_bert_config.min_clip_cosine_similarity_with_bert
                and clip_similarity.item()
                >= self.config.consider_annotation_with_bert_config.min_clip_cosine_similarity
            ):
                return True
        return False

    def inference(self, dataset: ImagesAndAnnotationsDataset) -> List[ImageAnnotation]:
        print(self.config)
        filtered_annotations: List[ImageAnnotation] = []
        image_number = 0
        for index, image_and_annotation in enumerate(dataset):
            annotation = ImageAnnotation()
            annotation.image_index = index
            for bounding_box_index, bounding_box in enumerate(
                image_and_annotation[1].bounding_boxes
            ):
                width = bounding_box.box.x2 - bounding_box.box.x1
                height = bounding_box.box.y2 - bounding_box.box.y1
                if width * height >= self.config.min_area and (
                    width >= self.config.min_width or height >= self.config.min_height
                ):
                    is_detected: bool = False
                    if (
                        self.config.classify_method
                        == ClipClassifyMethod.CLIP_WITH_LVIS_AND_CUSTOM_VOCABULARY
                    ):
                        is_detected = self.inference_with_lvis_vocabulary(bounding_box)
                    elif (
                        self.config.classify_method
                        == ClipClassifyMethod.CONSIDER_ANNOTATION_WITH_BERT
                    ):
                        is_detected = self.inference_with_bert(bounding_box)
                    else:
                        raise RuntimeError(
                            "Classify method "
                            + self.config.classify_method.value
                            + " does not support."
                        )
                    if is_detected:
                        # print("Image Number : " + str(image_number))
                        # print(
                        #     (
                        #         int(bounding_box.box.x1),
                        #         int(bounding_box.box.y1),
                        #         int(bounding_box.box.x2),
                        #         int(bounding_box.box.y2),
                        #     )
                        # )
                        # self.to_pil_image(copy.deepcopy(image_and_annotation[0])).crop(
                        #     (
                        #         int(bounding_box.box.x1),
                        #         int(bounding_box.box.y1),
                        #         int(bounding_box.box.x2),
                        #         int(bounding_box.box.y2),
                        #     )
                        # ).save("data/" + str(image_number) + ".jpeg")
                        # print("")
                        annotation.bounding_boxes.append(copy.deepcopy(bounding_box))
                        image_number = image_number + 1
        return filtered_annotations



from dataclasses import dataclass, field
from dataclass_wizard import YAMLWizard
from amber_mcap.exception import TaskDescriptionError
from enum import Enum
from typing import Dict
import os


class DeticModelType(Enum):
    SwinB_896_4x = "SwinB_896_4x"
    R50_640_4x = "R50_640_4x"


class DeticVocabulary(Enum):
    LVIS = "lvis"
    IMAGENET_21K = "imagenet_21k"


@dataclass
class DockerConfig(YAMLWizard):  # type: ignore
    use_gpu: bool = False
    claenup_image_on_shutdown: bool = False
    shm_size: str = "64M"


@dataclass
class DeticImageLabalerConfig(YAMLWizard):  # type: ignore
    model_type: DeticModelType = DeticModelType.SwinB_896_4x
    vocabulary: DeticVocabulary = DeticVocabulary.LVIS
    video_output_path: str = ""  # If the text is empty, it means no video output.
    min_height: int = 5
    min_width: int = 5
    min_area: int = 50

    def validate(self) -> None:
        if (
            self.video_output_path != ""
            and os.path.splitext(self.video_output_path)[-1] != ".mp4"
        ):
            raise TaskDescriptionError(
                "Type of the output video should be mp4, you specified "
                + self.video_output_path
            )

    def get_onnx_filename(self) -> str:
        if self.model_type == DeticModelType.SwinB_896_4x:
            if self.vocabulary == DeticVocabulary.LVIS:
                return "Detic_C2_SwinB_896_4x_IN-21K+COCO_lvis_op16.onnx"
            elif self.vocabulary == DeticVocabulary.IMAGENET_21K:
                return "Detic_C2_SwinB_896_4x_IN-21K+COCO_in21k_op16.onnx"
        elif self.model_type == DeticModelType.R50_640_4x:
            if self.vocabulary == DeticVocabulary.LVIS:
                return "Detic_C2_R50_640_4x_lvis_op16.onnx"
            elif self.vocabulary == DeticVocabulary.IMAGENET_21K:
                return "Detic_C2_R50_640_4x_in21k_op16.onnx"

        raise TaskDescriptionError(
            "Model type and vocablary is invalid, please check setting. You specified, model_type : "
            + self.model_type.value
            + " , vocabulary : ",
            self.vocablary.value,
        )


class ClipClassifyMethod(Enum):
    CLIP_WITH_LVIS_AND_CUSTOM_VOCABULARY = "clip_with_lvis_and_custom_vocabulary"
    CONSIDER_ANNOTATION_WITH_BERT = "consider_annotation_with_bert"


@dataclass
class ConsiderAnnotationWithBerfConfig(YAMLWizard):  # type: ignore
    positive_nagative_ratio: float = 2.0
    min_clip_cosine_similarity: float = 0.25
    min_clip_cosine_similarity_with_bert: float = 0.5


@dataclass
class ClipImageAnnotationFilterConfig(YAMLWizard):  # type: ignore
    target_objects: list[str] = field(default_factory=list)
    min_height: int = 5
    min_width: int = 5
    min_area: int = 50
    classify_method: ClipClassifyMethod = (
        ClipClassifyMethod.CONSIDER_ANNOTATION_WITH_BERT
    )
    consider_annotation_with_bert_config: ConsiderAnnotationWithBerfConfig = (
        ConsiderAnnotationWithBerfConfig()
    )


@dataclass
class ColmapPoseEstimationConfig(YAMLWizard):  # type: ignore
    docker_config: DockerConfig = DockerConfig()



class Schema:
    name: str
    schema_text: str


class StringMessageSchema(Schema):
    name = "std_msgs/msg/String"
    schema_text = """\
string data"""


class ImageMessageSchema(Schema):
    name = "sensor_msgs/msg/Image"
    schema_text = """\
std_msgs/Header header
uint32 height
uint32 width
string encoding
uint8 is_bigendian
uint32 step
uint8[] data
================================================================================
MSG: std_msgs/Header
builtin_interfaces/Time stamp
string frame_id"""


class CompressedMessageSchema(Schema):
    name = "sensor_msgs/msg/CompressedImage"
    schema_text = """\
std_msgs/Header header
string format
uint8[] data
================================================================================
MSG: std_msgs/Header
builtin_interfaces/Time stamp
string frame_id"""


class TFMessageSchema(Schema):
    name = "tf2_msgs/msg/TFMessage"
    schema_text = """geometry_msgs/TransformStamped[] transforms\n\n================================================================================\nMSG: geometry_msgs/TransformStamped\n# This expresses a transform from coordinate frame header.frame_id\n# to the coordinate frame child_frame_id at the time of header.stamp\n#\n# This message is mostly used by the\n# <a href="https://index.ros.org/p/tf2/">tf2</a> package.\n# See its documentation for more information.\n#\n# The child_frame_id is necessary in addition to the frame_id\n# in the Header to communicate the full reference for the transform\n# in a self contained message.\n\n# The frame id in the header is used as the reference frame of this transform.\nstd_msgs/Header header\n\n# The frame id of the child frame to which this transform points.\nstring child_frame_id\n\n# Translation and rotation in 3-dimensions of child_frame_id from header.frame_id.\nTransform transform\n\n================================================================================\nMSG: geometry_msgs/Transform\n# This represents the transform between two coordinate frames in free space.\n\nVector3 translation\nQuaternion rotation\n\n================================================================================\nMSG: geometry_msgs/Quaternion\n# This represents an orientation in free space in quaternion form.\n\nfloat64 x 0\nfloat64 y 0\nfloat64 z 0\nfloat64 w 1\n\n================================================================================\nMSG: geometry_msgs/Vector3\n# This represents a vector in free space.\n\n# This is semantically different than a point.\n# A vector is always anchored at the origin.\n# When a transform is applied to a vector, only the rotational component is applied.\n\nfloat64 x\nfloat64 y\nfloat64 z\n\n================================================================================\nMSG: std_msgs/Header\n# Standard metadata for higher-level stamped data types.\n# This is generally used to communicate timestamped data\n# in a particular coordinate frame.\n\n# Two-integer timestamp that is expressed as seconds and nanoseconds.\nbuiltin_interfaces/Time stamp\n\n# Transform frame with which this data is associated.\nstring frame_id\n\n================================================================================\nMSG: builtin_interfaces/Time\n# This message communicates ROS Time defined here:\n# https://design.ros2.org/articles/clock_and_time.html\n\n# The seconds component, valid over all int32 values.\nint32 sec\n\n# The nanoseconds component, valid in the range [0, 10e9).\nuint32 nanosec\n"""
    encoding = "ros2msg"



import os

from amber_mcap.dataset.images_dataset import ImagesDataset
from amber_mcap.automation.automation import Automation
from amber_mcap.automation.clip_encoder import ClipEncoder

import amber_mcap
from amber_mcap.automation.task_description import (
    DeticImageLabalerConfig,
)
from amber_mcap.automation.annotation import ImageAnnotation, BoundingBoxAnnotation


from amber_mcap.util.lvis.lvis_v1_categories import (
    LVIS_CATEGORIES as LVIS_V1_CATEGORIES,
)
from amber_mcap.util.imagenet_21k.in21k_categories import IN21K_CATEGORIES
from amber_mcap.util.color import color_brightness, random_color

import os
from torchvision import transforms
from tqdm import tqdm
from typing import Any, List, Dict, Tuple
import shutil
from download import download
import onnxruntime
from PIL import Image

# from torch.nn.functional import grid_sample
import numpy as np
import cv2
import json
import copy


class DeticImageLabeler(Automation):  # type: ignore
    def __init__(self, yaml_path: str) -> None:
        self.config = DeticImageLabalerConfig.from_yaml_file(yaml_path)
        self.weight_and_model = self.download_onnx(self.config.get_onnx_filename())
        self.session = onnxruntime.InferenceSession(
            self.weight_and_model,
            providers=["CPUExecutionProvider"],  # "CUDAExecutionProvider"],
        )
        self.to_pil_image = transforms.ToPILImage()

    def download_onnx(
        self,
        model: str,
        base_url: str = "https://storage.googleapis.com/ailia-models/detic/",
    ) -> str:
        download_directory = os.path.join(amber_mcap.__path__[0], "automation", "onnx")
        weight_path = os.path.join(download_directory, model)
        if not os.path.exists(weight_path):
            download(base_url + model, weight_path)
        return weight_path

    # This code comes from https://github.com/axinc-ai/ailia-models/blob/da1c277b602606586cd83943ef6b23eb705ec604/object_detection/detic/detic.py#L276-L301
    def preprocess(self, image: np.ndarray, detection_width: int = 800) -> np.ndarray:
        height, width, _ = image.shape
        image = image[:, :, ::-1]  # BGR -> RGB
        size = detection_width
        max_size = detection_width
        scale = size / min(height, width)
        if height < width:
            oh, ow = size, scale * width
        else:
            oh, ow = scale * height, size
        if max(oh, ow) > max_size:
            scale = max_size / max(oh, ow)
            oh = oh * scale
            ow = ow * scale
        ow = int(ow + 0.5)
        oh = int(oh + 0.5)
        image = np.asarray(Image.fromarray(image).resize((ow, oh), Image.BILINEAR))
        image = image.transpose((2, 0, 1))  # HWC -> CHW
        image = np.expand_dims(image, axis=0)
        image = image.astype(np.float32)
        return image

    # This function comes from https://github.com/axinc-ai/ailia-models/blob/da1c277b602606586cd83943ef6b23eb705ec604/object_detection/detic/detic.py#L153-L174
    def mask_to_polygons(self, mask: np.ndarray) -> List[Any]:
        # cv2.RETR_CCOMP flag retrieves all the contours and arranges them to a 2-level
        # hierarchy. External contours (boundary) of the object are placed in hierarchy-1.
        # Internal contours (holes) are placed in hierarchy-2.
        # cv2.CHAIN_APPROX_NONE flag gets vertices of polygons from contours.
        mask = np.ascontiguousarray(
            mask
        )  # some versions of cv2 does not support incontiguous arr
        res = cv2.findContours(
            mask.astype("uint8"), cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE
        )
        hierarchy = res[-1]
        if hierarchy is None:  # empty mask
            return []
        res = res[-2]
        res = [x.flatten() for x in res]
        # These coordinates from OpenCV are integers in range [0, W-1 or H-1].
        # We add 0.5 to turn them into real-value coordinate space. A better solution
        # would be to first +0.5 and then dilate the returned polygon by 0.5.
        return [x + 0.5 for x in res if len(x) >= 6]

    # This function comes from https://github.com/axinc-ai/ailia-models/blob/da1c277b602606586cd83943ef6b23eb705ec604/object_detection/detic/dataset_utils.py#L5-L11
    def get_lvis_meta_v1(self) -> Dict[str, List[str]]:
        thing_classes = [k["synonyms"][0] for k in LVIS_V1_CATEGORIES]
        meta = {"thing_classes": thing_classes}
        return meta

    # This function comes from https://github.com/axinc-ai/ailia-models/blob/da1c277b602606586cd83943ef6b23eb705ec604/object_detection/detic/dataset_utils.py#L14-L18
    def get_in21k_meta_v1(self) -> Dict[str, List[str]]:
        thing_classes = IN21K_CATEGORIES
        meta = {"thing_classes": thing_classes}
        return meta

    # This function comes from https://github.com/axinc-ai/ailia-models/blob/da1c277b602606586cd83943ef6b23eb705ec604/object_detection/detic/detic.py#L177-L269
    def draw_predictions(
        self, image: np.ndarray, detection_results: Any, vocabulary: str
    ) -> np.ndarray:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        height, width = image.shape[:2]

        boxes = detection_results["boxes"].astype(np.int64)
        scores = detection_results["scores"]
        classes = detection_results["classes"].tolist()
        masks = detection_results["masks"].astype(np.uint8)

        class_names = (
            self.get_lvis_meta_v1()
            if vocabulary == "lvis"
            else self.get_in21k_meta_v1()
        )["thing_classes"]
        labels = [class_names[i] for i in classes]
        labels = ["{} {:.0f}%".format(l, s * 100) for l, s in zip(labels, scores)]

        num_instances = len(boxes)

        np.random.seed()
        assigned_colors = [random_color(maximum=255) for _ in range(num_instances)]

        areas = np.prod(boxes[:, 2:] - boxes[:, :2], axis=1)
        if areas is not None:
            sorted_idxs = np.argsort(-areas).tolist()
            # Re-order overlapped instances in descending order.
            boxes = boxes[sorted_idxs]
            labels = [labels[k] for k in sorted_idxs]
            masks = [masks[idx] for idx in sorted_idxs]
            assigned_colors = [assigned_colors[idx] for idx in sorted_idxs]

        default_font_size = int(max(np.sqrt(height * width) // 90, 10))

        for i in range(num_instances):
            color = assigned_colors[i]
            color = (int(color[0]), int(color[1]), int(color[2]))
            image_b = image.copy()

            # draw box
            x0, y0, x1, y1 = boxes[i]
            cv2.rectangle(
                image_b,
                (x0, y0),
                (x1, y1),
                color=color,
                thickness=default_font_size // 4,
            )

            # draw segment
            polygons = self.mask_to_polygons(masks[i])
            for points in polygons:
                points = np.array(points).reshape((1, -1, 2)).astype(np.int32)
                cv2.fillPoly(image_b, pts=[points], color=color)

            image = cv2.addWeighted(image, 0.5, image_b, 0.5, 0)

        for i in range(num_instances):
            color = assigned_colors[i]
            color_text = color_brightness(color, brightness_factor=0.7)

            color = (int(color[0]), int(color[1]), int(color[2]))
            color_text = (int(color_text[0]), int(color_text[1]), int(color_text[2]))

            x0, y0, x1, y1 = boxes[i]

            SMALL_OBJECT_AREA_THRESH = 1000
            instance_area = (y1 - y0) * (x1 - x0)

            # for small objects, draw text at the side to avoid occlusion
            text_pos = (x0, y0)  # if drawing boxes, put text on the box corner.
            if instance_area < SMALL_OBJECT_AREA_THRESH or y1 - y0 < 40:
                if y1 >= height - 5:
                    text_pos = (x1, y0)
                else:
                    text_pos = (x0, y1)

            # draw label
            x, y = text_pos
            text = labels[i]
            font = cv2.FONT_HERSHEY_SIMPLEX
            height_ratio = (y1 - y0) / np.sqrt(height * width)
            font_scale = np.clip((height_ratio - 0.02) / 0.08 + 1, 1.2, 2) * 0.5
            font_thickness = 1
            text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
            text_w, text_h = text_size
            cv2.rectangle(
                image, text_pos, (int(x + text_w * 0.6), y + text_h), (0, 0, 0), -1
            )
            cv2.putText(
                image,
                text,
                (x, y + text_h - 5),
                fontFace=font,
                fontScale=font_scale * 0.6,
                color=color_text,
                thickness=font_thickness,
                lineType=cv2.LINE_AA,
            )

        return image

    def inference(self, dataset: ImagesDataset) -> List[ImageAnnotation]:
        clip_encoder = ClipEncoder()
        image_annotations: List[ImageAnnotation] = []
        vocabulary = "lvis"
        class_names = (
            self.get_lvis_meta_v1()
            if vocabulary == "lvis"
            else self.get_in21k_meta_v1()
        )["thing_classes"]

        video: Any = None
        for index, image in enumerate(dataset):
            print("Labeling image " + str(index) + "/" + str(len(dataset)))
            input_width = image.shape[2]
            input_height = image.shape[1]
            input_image = self.preprocess(np.asarray(self.to_pil_image(image)), 800)
            boxes, scores, classes, masks = self.session.run(
                None,
                {
                    "img": input_image,
                    "im_hw": np.array([input_height, input_width]).astype(np.int64),
                },
            )
            image_annotation = ImageAnnotation()
            image_annotation.image_index = index
            detection_results = {
                "boxes": boxes,
                "scores": scores,
                "classes": classes,
                "masks": masks,
            }
            for bbox_id in range(len(boxes)):
                bounding_box = BoundingBoxAnnotation()
                bounding_box.box.x1 = boxes[bbox_id][0]
                bounding_box.box.y1 = boxes[bbox_id][1]
                bounding_box.box.x2 = boxes[bbox_id][2]
                bounding_box.box.y2 = boxes[bbox_id][3]
                bounding_box.score = scores[bbox_id]
                bounding_box.object_class = class_names[classes[bbox_id]]
                if (
                    abs(bounding_box.box.x2 - bounding_box.box.x1)
                    >= self.config.min_width
                    or abs(bounding_box.box.y2 - bounding_box.box.y1)
                    >= self.config.min_height
                ) and abs(bounding_box.box.x2 - bounding_box.box.x1) * abs(
                    bounding_box.box.y2 - bounding_box.box.y1
                ) >= self.config.min_area:
                    image_annotation.bounding_boxes.append(copy.deepcopy(bounding_box))
            image_annotations.append(
                clip_encoder.get_single_image_embeddings_for_objects(
                    self.to_pil_image(image), image_annotation
                )
            )
            visualization = self.draw_predictions(
                np.asarray(self.to_pil_image(image)), detection_results, "lvis"
            )
            if video == None:
                video = cv2.VideoWriter(
                    self.config.video_output_path,
                    cv2.VideoWriter_fourcc("m", "p", "4", "v"),
                    30.0,  # FPS
                    (visualization.shape[1], visualization.shape[0]),
                )
            video.write(visualization)
        video.release()
        return image_annotations



from amber_mcap.dataset.conversion import build_transform_stamped_message
from amber_mcap.dataset.rosbag2_dataset import Rosbag2Dataset, MessageMetaData
from amber_mcap.dataset.topic_config import TfTopicConfig
from amber_mcap.unit.time import Time, TimeUnit
from dataclass_wizard import YAMLWizard
from dataclasses import dataclass
from mcap.reader import NonSeekingReader
from tf2_amber import BufferCore, timeFromSec, durationFromSec, TransformStamped
import sys
import torch
from typing import List
import datetime


@dataclass
class ReadTfTopicConfig(YAMLWizard):  # type: ignore
    tf_topic: TfTopicConfig = TfTopicConfig()
    compressed: bool = True
    sampling_duration: float = 0.01666666666  # 60Hz
    target_frame: str = ""
    source_frame: str = ""

    def get_tf_topic(self) -> str:
        return self.tf_topic.topic_name

    def get_static_tf_topic(self) -> str:
        return self.tf_topic.static_tf_topic_name


class TfDataset(Rosbag2Dataset):  # type: ignore
    def __init__(
        self,
        rosbag_path: str,
        config: ReadTfTopicConfig,
    ) -> None:
        self.config = config
        self.transforms: List[TransformStamped] = []
        print(self.config)
        super().__init__(
            rosbag_path,
            self.config.compressed,
        )
        first_timestamp = None
        last_timestamp = None
        tf_buffer = BufferCore(durationFromSec(sys.float_info.max))
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                timestamp = message.log_time

                if first_timestamp is None or timestamp < first_timestamp:
                    first_timestamp = timestamp
                if last_timestamp is None or timestamp > last_timestamp:
                    last_timestamp = timestamp

                if channel.topic in [self.config.get_tf_topic()]:
                    for tf_amber_message in build_transform_stamped_message(
                        message, schema, self.config.compressed
                    ):
                        tf_buffer.setTransform(
                            tf_amber_message, "Authority undetectable", False
                        )
                if channel.topic in [self.config.get_static_tf_topic()]:
                    for tf_amber_message in build_transform_stamped_message(
                        message, schema, self.config.compressed
                    ):
                        tf_buffer.setTransform(
                            tf_amber_message, "Authority undetectable", True
                        )
        sampled_timestamps = self.get_sampled_timestamps(
            Time(float(first_timestamp), TimeUnit.NANOSECOND),
            Time(float(last_timestamp), TimeUnit.NANOSECOND),
            Time(self.config.sampling_duration, TimeUnit.SECOND),
        )
        for timestamp in sampled_timestamps:
            timestamp_nanosec = Time(float(timestamp), TimeUnit.NANOSECOND)
            try:
                transform = tf_buffer.lookupTransform(
                    self.config.source_frame,
                    self.config.target_frame,
                    timeFromSec(timestamp_nanosec.get(TimeUnit.SECOND)),
                )
                self.transforms.append(transform)
            except:
                pass
            self.message_metadata.append(
                MessageMetaData.from_dict(
                    {
                        "publish_time": datetime.datetime.fromtimestamp(
                            timestamp_nanosec.get(TimeUnit.SECOND),
                            tz=datetime.timezone.utc,
                        ),
                        "topic": "N/A",
                        "rosbag_path": "N/A",
                    }
                )
            )

    def __len__(self) -> int:
        return len(self.transforms)

    def __iter__(self) -> torch.Tensor:
        current_index = 0
        for transfrom in self.transforms:
            current_index = current_index + 1
            yield torch.Tensor(
                [
                    transfrom.transform.translation.x,
                    transfrom.transform.translation.y,
                    transfrom.transform.translation.z,
                    transfrom.transform.rotation.x,
                    transfrom.transform.rotation.y,
                    transfrom.transform.rotation.z,
                    transfrom.transform.rotation.w,
                ]
            )

    def get_sampled_timestamps(
        self, first_timestamp: Time, last_timestamp: Time, sampling_duration: Time
    ) -> List[int]:
        return list(
            range(
                int(first_timestamp.get(TimeUnit.NANOSECOND)),
                int(last_timestamp.get(TimeUnit.NANOSECOND)),
                int(sampling_duration.get(TimeUnit.NANOSECOND)),
            )
        )



from dataclasses import dataclass
from dataclass_wizard import YAMLWizard
from enum import Enum


@dataclass
class ImageTopicConfig(YAMLWizard):  # type: ignore
    topic_name: str = ""
    compressed_image: bool = False


class PointCloudType(Enum):
    XYZ = "XYZ"


@dataclass
class PointcloudTopicConfig(YAMLWizard):  # type: ignore
    topic_name: str = ""
    pointcloud_type: PointCloudType = PointCloudType.XYZ


@dataclass
class TfTopicConfig(YAMLWizard):  # type: ignore
    topic_name: str = "/tf"
    static_tf_topic_name: str = "/tf_static"



from amber_mcap.dataset.rosbag2_dataset import Rosbag2Dataset, MessageMetaData
from amber_mcap.dataset.topic_config import PointcloudTopicConfig
import torch
from dataclasses import dataclass, field
from dataclass_wizard import YAMLWizard
from amber_mcap.unit.time import Time, TimeUnit
from amber_mcap.dataset.conversion import decode_pointcloud_message
from typing import Any, List, Dict
from amber_mcap.dataset.rosbag2_dataset import Rosbag2Dataset
from mcap.reader import NonSeekingReader
import open3d
import datetime


@dataclass
class ReadPointCloudConfig(YAMLWizard):  # type: ignore
    pointcloud_topics: List[PointcloudTopicConfig] = field(default_factory=list)
    compressed: bool = True

    def get_pointcloud_topics(self) -> List[str]:
        topics: List[str] = []
        for topic in self.pointcloud_topics:
            topics.append(topic.topic_name)
        return topics


class PointcloudDataset(Rosbag2Dataset):  # type: ignore
    num_pointclouds = 0
    config: ReadPointCloudConfig = ReadPointCloudConfig()

    def __init__(
        self,
        rosbag_path: str,
        config: ReadPointCloudConfig,
        transform: Any = None,
        target_transform: Any = None,
    ) -> None:
        self.config = config
        print(self.config)
        super().__init__(
            rosbag_path,
            self.config.compressed,
            transform,
            target_transform,
        )
        self.count_pointclouds()

    def count_pointclouds(self) -> None:
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                if channel.topic in self.config.get_pointcloud_topics():
                    self.num_pointclouds = self.num_pointclouds + 1
                    self.message_metadata.append(
                        MessageMetaData.from_dict(
                            {
                                "publish_time": datetime.datetime.fromtimestamp(
                                    Time(message.publish_time, TimeUnit.NANOSECOND).get(
                                        TimeUnit.SECOND
                                    ),
                                    tz=datetime.timezone.utc,
                                ),
                                "topic": channel.topic,
                                "rosbag_path": rosbag_file,
                            }
                        )
                    )

    def __len__(self) -> int:
        return self.num_pointclouds

    def __iter__(self) -> torch.Tensor:
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                if channel.topic in self.config.get_pointcloud_topics():
                    yield decode_pointcloud_message(
                        message, schema, self.config.compressed
                    )



from amber_mcap.automation.blip2_encoder import Blip2Encoder
from amber_mcap.automation.clip_encoder import ClipEncoder
from amber_mcap.dataset.images_dataset import ImagesDataset, ReadImagesConfig
from glob import glob
from pathlib import Path
from qdrant_client import QdrantClient
from qdrant_client.http.models import Distance, VectorParams
from qdrant_client.http.models import PointStruct
from typing import List, Any, Optional, cast
import os
import torch
import torchvision
import uuid
import docker
import gradio as gr
import hashlib
import argparse
from tqdm import tqdm
from torchvision import transforms


class SearchResult:
    duration_from_rosbag_start: float = 0
    mcap_path: str = ""
    image_path: str = ""

    def __init__(
        self, duration_from_rosbag_start: float, mcap_path: str, image_path: str
    ) -> None:
        self.duration_from_rosbag_start = duration_from_rosbag_start
        self.mcap_path = mcap_path
        self.image_path = image_path


class ImageSearch:
    def __init__(
        self,
        qdrant_port: int = 5555,
        preload_rosbag_directory: Optional[str] = None,
        model: str = "blip2",
    ) -> None:
        # make data saving directory
        self.data_directory = "/tmp/image_search"
        self.mcap_hashes: List[str] = []
        if not os.path.exists(self.data_directory):
            os.makedirs(self.data_directory)

        # setup docker container
        container_name = "image_search"
        self.docker_client = docker.from_env()
        if not self.found_image(image_name="qdrant/qdrant", tag="v1.7.2"):
            self.docker_client.images.pull("qdrant/qdrant", tag="v1.7.2")
        if not self.found_container(container_name):
            self.container = self.docker_client.containers.run(
                image="qdrant/qdrant:v1.7.2",
                ports={6333: qdrant_port},
                detach=True,
                tty=True,
                name=container_name,
            )

        # setup qdrant client
        self.client = QdrantClient("localhost", port=qdrant_port)
        self.model = model
        if self.model == "blip2":
            # load blip2 models
            self.encoder = Blip2Encoder()
        elif self.model == "clip":
            self.encoder = ClipEncoder()
        else:
            raise Exception("Model type " + self.model + " does not supported.")

        if preload_rosbag_directory != None:
            self.preload_rosbag_files(Path(str(preload_rosbag_directory)))

    def collection_exists(self, collection_name: str) -> bool:
        for collections in self.client.get_collections():
            for collection in collections[1]:
                if collection.name == collection_name:
                    return True
        return False

    def is_processed_mcap(self, mcap_path: str) -> bool:
        hash = hashlib.sha256()
        with open(mcap_path, "rb") as f:
            while True:
                chunk = f.read(2048 * hash.block_size)
                if len(chunk) == 0:
                    break
                hash.update(chunk)
        digest = hash.hexdigest()
        if digest in self.mcap_hashes:
            return True
        else:
            self.mcap_hashes.append(digest)
            return False

    def show_gradio_ui(self) -> None:
        with gr.Blocks() as gradio_ui:
            chatbot = gr.Chatbot()
            msg = gr.Textbox()
            clear = gr.ClearButton([msg, chatbot])
            upload_button = gr.UploadButton(
                "Click to upload a directory which contains rosbag in mcap format and dataset.yaml",
                file_count="directory",
            )
            file_output = gr.File()
            upload_button.upload(self.upload_file, upload_button, file_output)

            def respond(message: Any, chat_history: Any) -> Any:
                result = self.search_by_text(message)
                if result == None:
                    chat_history.append(
                        (message, ("something wrong happend in search"))
                    )
                else:
                    chat_history.append(
                        (message, (cast(SearchResult, result).image_path,))
                    )
                    chat_history.append(
                        ("Which rosbag?", cast(SearchResult, result).mcap_path)
                    )
                    chat_history.append(
                        (
                            "Around how many seconds after the start?",
                            "About "
                            + str(cast(SearchResult, result).duration_from_rosbag_start)
                            + " sec",
                        )
                    )
                return "", chat_history

            msg.submit(respond, [msg, chatbot], [msg, chatbot])
        gradio_ui.launch()

    def upload_file(self, files: List[Any]) -> List[str]:
        mcap_path: Optional[str] = None
        yaml_path: Optional[str] = None
        for file in files:
            if os.path.splitext(file.name)[-1] == ".mcap":
                if mcap_path != None:
                    return []
                mcap_path = file.name
            if os.path.basename(file.name) == "dataset.yaml":
                if yaml_path != None:
                    return []
                yaml_path = file.name
        if mcap_path == None or yaml_path == None:
            return []

        if self.is_processed_mcap(str(mcap_path)):
            return []
        self.preprocess(
            ImagesDataset(mcap_path, ReadImagesConfig.from_yaml_file(yaml_path))
        )
        return [str(mcap_path), str(yaml_path)]

    def found_container(self, container_name: str) -> bool:
        for container in self.docker_client.containers.list():
            if container.name == container_name:
                return True
        return False

    def found_image(self, image_name: str, tag: str = "latest") -> bool:
        qdrant_images = self.docker_client.images.list(
            filters={"reference": image_name}
        )
        for qdrant_image in qdrant_images:
            for docker_tag in qdrant_image.tags:
                if docker_tag == image_name + ":" + tag:
                    return True
        return False

    def search_by_text(self, text: str) -> Optional[SearchResult]:
        if self.model == "blip2":
            search_result = self.client.search(
                collection_name=self.model,
                query_vector=self.encoder.encode_text(text)[0].tolist(),
                limit=10,
            )
        elif self.model == "clip":
            search_result = self.client.search(
                collection_name=self.model,
                query_vector=self.encoder.get_single_text_embeddings(text)[0].tolist(),
                limit=10,
            )
        else:
            raise Exception("Model type " + self.model + " does not supported.")
        for result in search_result:
            return SearchResult(
                mcap_path=result.payload["mcap_path"],
                image_path=result.payload["image_path"],
                duration_from_rosbag_start=result.payload["duration_from_rosbag_start"],
            )
        return None

    def preload_rosbag_files(self, rosbag_directory: Path) -> None:
        for mcap_file in tqdm(
            glob(
                "**/*.mcap",
                root_dir=rosbag_directory.absolute().as_posix(),
                recursive=True,
            ),
            desc="preloading rosbag",
            postfix="range",
            ncols=80,
        ):
            print(
                "Start loading rosbag : "
                + os.path.join(rosbag_directory.absolute().as_posix(), mcap_file)
            )
            if not self.is_processed_mcap(
                os.path.join(rosbag_directory.absolute().as_posix(), mcap_file)
            ):
                self.preprocess(
                    ImagesDataset(
                        os.path.join(rosbag_directory.absolute().as_posix(), mcap_file),
                        ReadImagesConfig.from_yaml_file(
                            os.path.join(
                                rosbag_directory.absolute().as_posix(), "dataset.yaml"
                            )
                        ),
                    )
                )
            else:
                print("Rosbag data was already processed.")

    def preprocess(self, dataset: ImagesDataset) -> None:
        if not self.collection_exists(self.model):
            if self.model == "blip2":
                self.client.create_collection(
                    collection_name=self.model,
                    vectors_config=VectorParams(size=256, distance=Distance.COSINE),
                )
            else:
                self.client.create_collection(
                    collection_name=self.model,
                    vectors_config=VectorParams(size=512, distance=Distance.COSINE),
                )
        batch_size = 1
        dataloader = torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, shuffle=False, num_workers=0
        )
        for i_batch, sample_batched in enumerate(dataloader):
            for sample_index, sample in enumerate(
                tqdm(sample_batched, desc="processing images in rosbag.")
            ):
                image_features: torch.Tensor
                if self.model == "blip2":
                    image_features = self.encoder.encode_image(sample)
                elif self.model == "clip":
                    to_pil_image = transforms.ToPILImage()
                    image_features = self.encoder.get_single_image_embeddings(
                        to_pil_image(sample)
                    ).unsqueeze(0)
                else:
                    raise Exception("Model type " + self.model + " does not supported.")
                points = []
                image_id = uuid.uuid4()
                image_path = os.path.join(self.data_directory, str(image_id) + ".png")
                torchvision.transforms.functional.to_pil_image(sample).save(image_path)
                for i in range(image_features[0].shape[0]):
                    points.append(
                        PointStruct(
                            id=str(uuid.uuid4()),
                            vector=image_features[0][i].tolist(),
                            payload={
                                "image_path": image_path,
                                "image_id": image_id,
                                "mcap_path": dataset.rosbag_files,
                                "duration_from_rosbag_start": (
                                    dataset.get_metadata(
                                        i_batch * batch_size + sample_index
                                    ).publish_time
                                    - dataset.get_first_timestamp()
                                ).total_seconds(),
                            },
                        )
                    )
                    self.client.upsert(
                        collection_name=self.model, wait=True, points=points
                    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Sample application of searching image by blip2"
    )
    parser.add_argument("--port", default=5555, help="connection port of the qdrant")
    parser.add_argument("--model", choices=["blip2", "clip"], default="blip2")
    parser.add_argument("--rosbag_directory", default=None)
    args = parser.parse_args()
    app = ImageSearch(
        qdrant_port=args.port,
        preload_rosbag_directory=args.rosbag_directory,
        model=args.model,
    )
    app.show_gradio_ui()



from amber_mcap.automation.clip_encoder import ClipEncoder
from PIL import Image
from amber_mcap.automation.annotation import ImageAnnotation, BoundingBoxAnnotation
from typing import List
from amber_mcap.dataset.images_and_annotations_dataset import (
    ImagesAndAnnotationsDataset,
)
import torch
from torchvision import transforms
import tensorboardX

from dataclasses import dataclass, field
from dataclass_wizard import YAMLWizard
from enum import Enum
import random


class VisualizationTarget(Enum):
    IMAGE_EMBEDDINGS = "image_embeddings"


@dataclass
class ClipEmbeddingVisualizationConfig(YAMLWizard):  # type: ignore
    label_image_size: int = 50
    max_visualization_items: int = 1000
    target: VisualizationTarget = VisualizationTarget.IMAGE_EMBEDDINGS
    custom_vocabulary: List[str] = field(default_factory=list)


class ClipEmbeddingsVisualization:
    def __init__(self, yaml_path: str) -> None:
        self.config = ClipEmbeddingVisualizationConfig.from_yaml_file(yaml_path)
        self.encoder = ClipEncoder()
        self.image_embeddings = torch.zeros(0)
        self.text_embeddings = torch.zeros(0)
        self.label_images = torch.zeros(0)
        self.transform = transforms.Compose(
            [
                transforms.Resize(
                    (self.config.label_image_size, self.config.label_image_size)
                ),
                transforms.ToTensor(),
            ]
        )
        self.current_object_index: int = 0
        self.to_pil_image = transforms.ToPILImage()

    def add_image_embedding(
        self, image: torch.Tensor, annotation: ImageAnnotation
    ) -> None:
        with torch.no_grad():
            pil_image = self.to_pil_image(image)
            for bounding_box in annotation.bounding_boxes:
                if self.current_object_index in self.target_object_index:
                    cropped_image = pil_image.crop(
                        (
                            int(bounding_box.box.x1),
                            int(bounding_box.box.y1),
                            int(bounding_box.box.x2),
                            int(bounding_box.box.y2),
                        )
                    )
                    self.image_embeddings = torch.cat(
                        (
                            self.image_embeddings,
                            self.encoder.get_single_image_embeddings(cropped_image),
                        )
                    )
                    self.label_images = torch.cat(
                        (self.label_images, self.transform(cropped_image))
                    )
                self.current_object_index = self.current_object_index + 1

    def add_text_embedding(self, object_name: str) -> None:
        self.text_embeddings = torch.cat(
            (
                self.text_embeddings,
                self.encoder.get_single_text_embeddings("A photo of a " + object_name),
            )
        )

    def visualize(self, dataset: ImagesAndAnnotationsDataset) -> None:
        self.num_objects = 0
        for index, image_and_annotation in enumerate(dataset):
            for bounding_box in image_and_annotation[1].bounding_boxes:
                self.num_objects = self.num_objects + 1
        if self.num_objects >= self.config.max_visualization_items:
            self.target_object_index = random.sample(
                list(range(self.num_objects)), self.config.max_visualization_items
            )
        else:
            self.target_object_index = list(range(self.num_objects))
        self.current_object_index = 0
        for index, image_and_annotation in enumerate(dataset):
            self.add_image_embedding(image_and_annotation[0], image_and_annotation[1])
        for object_name in self.config.custom_vocabulary:
            self.add_text_embedding(object_name)
        # Append Black images for text embeddings
        self.label_images = torch.cat(
            (
                self.label_images,
                torch.zeros(
                    (
                        3 * len(self.config.custom_vocabulary),
                        self.config.label_image_size,
                        self.config.label_image_size,
                    )
                ),
            )
        )
        image_metadata = [
            "image_in_rosbag" for i in range(len(self.target_object_index))
        ]
        embeddings = torch.cat((self.image_embeddings, self.text_embeddings))
        writer = tensorboardX.SummaryWriter()
        writer.add_embedding(
            embeddings.view(
                len(self.target_object_index) + len(self.config.custom_vocabulary), 512
            ),
            label_img=self.label_images.view(
                len(self.target_object_index) + len(self.config.custom_vocabulary),
                3,
                self.config.label_image_size,
                self.config.label_image_size,
            ),
            metadata=image_metadata + self.config.custom_vocabulary,
        )
        writer.close()



import cv2
from amber_mcap.dataset.schema import ImageMessageSchema
from amber_mcap.unit.time import Time, TimeUnit
from amber_mcap.dataset.conversion import build_message_from_image
from mcap_ros2.writer import Writer as McapWriter
from tqdm import tqdm
import math
from typing import Callable

from dataclasses import dataclass
from dataclass_wizard import YAMLWizard


@dataclass
class VideoImporterConfig(YAMLWizard):  # type: ignore
    topic_name: str = "/camera/image_raw"
    rosbag_path: str = "output.mcap"


class VideoImporter:
    def __init__(self, video_path: str, config: VideoImporterConfig) -> None:
        self.config: VideoImporterConfig = config
        self.capture = cv2.VideoCapture(video_path)

    def write(self) -> None:
        with open(self.config.rosbag_path, "wb") as f:
            writer = McapWriter(f)
            schema = writer.register_msgdef(
                ImageMessageSchema.name, ImageMessageSchema.schema_text
            )
            get_nanoseconds_timestamp: Callable[[], int] = lambda: math.floor(
                Time(self.capture.get(cv2.CAP_PROP_POS_MSEC), TimeUnit.MILLISECOND).get(
                    TimeUnit.NANOSECOND
                )
            )
            for i in tqdm(range(int(self.capture.get(cv2.CAP_PROP_FRAME_COUNT)))):
                ret, cv_image = self.capture.read()
                writer.write_message(
                    topic=self.config.topic_name,
                    schema=schema,
                    message=build_message_from_image(
                        cv_image,
                        "camera",
                        Time(
                            self.capture.get(cv2.CAP_PROP_POS_MSEC),
                            TimeUnit.MILLISECOND,
                        ),
                        "rgb8",
                    ),
                    log_time=get_nanoseconds_timestamp(),
                    publish_time=get_nanoseconds_timestamp(),
                    sequence=i,
                )
            writer.finish()



from amber_mcap.dataset.schema import TFMessageSchema
from amber_mcap.dataset.conversion import build_message_from_tf
from tf2_amber import TransformStamped
from mcap_ros2.writer import Writer as McapWriter
from dataclasses import dataclass
from dataclass_wizard import YAMLWizard
import math


@dataclass
class TfImporterConfig(YAMLWizard):  # type: ignore
    rosbag_path: str = "output.mcap"


class TfImporter:
    def __init__(self, config):
        self.config: TfImporterConfig = config
        self.file = open(self.config.rosbag_path, "wb")
        self.writer = McapWriter(self.file)
        self.schema = self.writer.register_msgdef(
            TFMessageSchema.name, TFMessageSchema.schema_text
        )
        self.index = 0

    # self.writer.finish throw error when we use it in destructor.
    def finish(self):
        self.writer.finish()
        self.file.close()

    def write(self, transform: TransformStamped):
        nanoseconds = int(
            transform.header.stamp.sec * math.pow(10, 9)
            + transform.header.stamp.nanosec
        )
        self.writer.write_message(
            topic="/tf",
            schema=self.schema,
            message=build_message_from_tf([transform]),
            log_time=nanoseconds,
            publish_time=nanoseconds,
            sequence=self.index,
        )
        self.index = self.index + 1


if __name__ == "__main__":
    pass
    # importer = TfImporter(TfImporterConfig())
    # importer.write(TransformStamped())
    # importer.write(TransformStamped())
    # importer.write(TransformStamped())
    # importer.finish()



from sentence_transformers import util, SentenceTransformer
from typing import Dict
import torch


class TextEncoder:
    def __init__(self, model_name: str = "all-MiniLM-L6-v2") -> None:
        self.model: SentenceTransformer = SentenceTransformer(model_name)
        self.cache: Dict[str, torch.Tensor] = {}

    def encode(self, text: str) -> torch.Tensor:
        if text in self.cache:
            return self.cache[text]
        self.cache[text] = self.model.encode(text)
        return self.cache[text]

    def cosine_similarity(self, text0: str, text1: str) -> float:
        return float(
            util.pytorch_cos_sim(self.encode(text0), self.encode(text1)).item()
        )



class TaskDescriptionError(Exception):
    pass


class MessageDecodingError(Exception):
    pass


class RosbagSchemaError(Exception):
    pass


class RuntimeError(Exception):
    pass


class CertificationError(Exception):
    pass



from typing import Any
from abc import ABC, abstractmethod
from amber_mcap.dataset.images_dataset import Rosbag2Dataset
from typing import List, Dict
import json
from mcap.reader import NonSeekingReader
from mcap_ros2.writer import Writer
from mcap.records import Schema
from amber_mcap.dataset.schema import StringMessageSchema
from amber_mcap.dataset.conversion import decode_message
from amber_mcap.exception import RosbagSchemaError
from mcap_ros2.decoder import Decoder
import sys
import datetime
from amber_mcap.unit.time import Time, TimeUnit


class Automation(ABC):
    @abstractmethod
    def __init__(self, yaml_path: str) -> None:
        pass

    @abstractmethod
    def inference(self, dataset: Rosbag2Dataset) -> Any:
        pass

    def write(
        self,
        dataset: Rosbag2Dataset,
        topic: str,
        annotation_data: List[Any],
        output_rosbag_path: str,
    ) -> None:
        annotation_json: List[str] = []
        rosbag_file = open(output_rosbag_path, "w+b")
        writer = Writer(output=rosbag_file)
        schema_dicts: Dict[str, Schema] = {}  # {schena name : schema}
        for rosbag_filepath in dataset.rosbag_files:
            reader = NonSeekingReader(rosbag_filepath)
            # Copy all topics
            for schema, channel, message in reader.iter_messages():
                if not schema.name in schema_dicts:
                    schema_dicts[schema.name] = writer.register_msgdef(
                        schema.name, schema.data.decode("utf-8")
                    )
                writer.write_message(
                    topic=channel.topic,
                    schema=schema_dicts[schema.name],
                    message=decode_message(message, schema, dataset.compressed),
                    log_time=message.log_time,
                    publish_time=message.publish_time,
                    sequence=message.sequence,
                )
        for annotation in annotation_data:
            annotation_json.append(annotation.to_json())
        # Append annotation data
        annotation_schema = writer.register_msgdef(
            StringMessageSchema.name, StringMessageSchema.schema_text
        )
        annotation_schema.id = len(schema_dicts) + 1
        stamp: int = int(
            Time(
                (
                    dataset.get_first_timestamp()
                    - datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
                ).total_seconds(),
                TimeUnit.SECOND,
            ).get(TimeUnit.NANOSECOND)
        )
        writer.write_message(
            topic=topic,
            schema=annotation_schema,
            message={"data": json.dumps(annotation_json)},
            sequence=0,
            publish_time=stamp,
            log_time=stamp,
        )
        writer.finish()
        rosbag_file.close()



import docker
from amber_mcap.automation.automation import Automation
import os
from amber_mcap.dataset.images_dataset import ImagesDataset
from amber_mcap.automation.task_description import ColmapPoseEstimationConfig
from typing import List
from torchvision import transforms
import uuid
import subprocess


class Nerf3DReconstruction(Automation):  # type: ignore
    docker_image_name = "dromni/nerfstudio:0.3.1"

    def __init__(self, yaml_path: str) -> None:
        self.config = ColmapPoseEstimationConfig.from_yaml_file(yaml_path)
        self.result_id = str(uuid.uuid4())
        self.temporary_directory = os.path.join(
            "/tmp/nerf_3d_reconstruction", self.result_id
        )
        self.setup_directory()
        self.to_pil_image = transforms.ToPILImage()
        self.docker_client = docker.from_env()
        self.docker_client.images.pull("dromni/nerfstudio", tag="0.3.1")
        self.container = self.docker_client.containers.run(
            image="dromni/nerfstudio:0.3.1",
            volumes={
                self.get_input_directory_path(): {
                    "bind": "/workspace/inputs",
                    "mode": "rw",
                },
            },
            # ports={7007: 7007},
            device_requests=self.build_device_requests(),
            shm_size=self.config.docker_config.shm_size,
            command=["/bin/bash"],
            detach=True,
            tty=True,
            runtime=None,
        )

    def get_input_directory_path(self) -> str:
        return os.path.join(self.temporary_directory, "inputs")

    def get_output_directory_path(self) -> str:
        return os.path.join(self.temporary_directory, "outputs")

    def get_camera_pose_output_directory_path(self) -> str:
        return os.path.join(self.get_output_directory_path(), "camera_poses")

    def get_checkpoint_output_directory_path(self) -> str:
        return os.path.join(self.get_output_directory_path(), "checkpoints")

    def get_trained_model_output_directory_path(self) -> str:
        return os.path.join(self.get_output_directory_path(), "models")

    def __del__(self) -> None:
        self.container.stop()
        self.container.remove()
        if self.config.docker_config.claenup_image_on_shutdown:
            self.docker_client.images.remove(
                image=self.docker_image_name, force=False, noprune=False
            )

    def build_device_requests(self) -> list[docker.types.DeviceRequest]:
        requests = []
        if self.config.docker_config.use_gpu:
            requests.append(
                docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
            )
        return requests

    def setup_directory(self) -> None:
        os.makedirs(self.temporary_directory)
        os.makedirs(self.get_input_directory_path())

    def cpu_or_gpu(self) -> str:
        if self.config.docker_config.use_gpu:
            return "--gpu"
        else:
            return "--no-gpu"

    def build_docker_copy_command(self) -> List[str]:
        return [
            "docker",
            "cp",
            self.container.id + ":/workspace/outputs",
            self.get_output_directory_path(),
        ]

    def build_post_process_command(self) -> List[str]:
        """Same as "ns-process-data images --data /workspace/inputs --output-dir /workspace/outputs/camera_poses"

        Returns:
            List[str]: bash command for post-processing data.
        """
        return [
            "ns-process-data",
            "images",
            "--data",
            "/workspace/inputs",
            "--output-dir",
            "/workspace/outputs/camera_poses",
            self.cpu_or_gpu(),
        ]

    def build_train_command(self) -> List[str]:
        """Same as "ns-train nerfact --data /workspace/outputs/camera_poses
            --output-dir /workspace/outputs/checkpoints
            --vis tensorboard --pipeline.model.predict-normals True"

        Returns:
            List[str]: _description_
        """
        return [
            "ns-train",
            "nerfacto",
            "--data",
            "/workspace/outputs/camera_poses",
            "--output-dir",
            "/workspace/outputs/checkpoints",
            "--vis",
            "tensorboard",
            "--pipeline.model.predict-normals",
            "True",
        ]

    def build_export_pointcloud_command(self) -> List[str]:
        """Same as "find -name config.yml | xargs -I {} ns-export pointcloud
            --load-config {} --output-dir /workspace/outputs/pointcloud"

        Returns:
            List[str]: bash command for exporting point cloud
        """
        return [
            "find",
            "-name",
            "config.yml",
            "|",
            "xargs",
            "-I",
            "{}",
            "ns-export",
            "pointcloud",
            "--load-config",
            "{}",
            "--output-dir",
            "/workspace/outputs/pointcloud",
        ]

    def build_export_mesh_command(self) -> List[str]:
        """Same as "find -name config.yml | xargs -I {} ns-export poisson --load-config {} --output-dir /workspace/outputs/mesh"

        Returns:
            List[str]: bash command for exporting mesh
        """
        return [
            "find",
            "-name",
            "config.yml",
            "|",
            "xargs",
            "-I",
            "{}",
            "ns-export",
            "poisson",
            "--load-config",
            "{}",
            "--output-dir",
            "/workspace/outputs/mesh",
        ]

    def run_command(self, command: List[str]) -> None:
        _, stream = self.container.exec_run(command, stream=True)
        for data in stream:
            print(data.decode())

    def inference(self, dataset: ImagesDataset) -> None:
        for index, image in enumerate(dataset):
            self.to_pil_image(image).save(
                os.path.join(
                    self.get_input_directory_path(),
                    "input" + str(index) + ".jpeg",
                )
            )
        self.run_command(self.build_post_process_command())
        self.run_command(self.build_train_command())
        self.run_command(self.build_export_pointcloud_command())
        self.run_command(self.build_export_mesh_command())
        subprocess.call(self.build_docker_copy_command())
        print("Artifacts are outputed under " + self.get_output_directory_path())
        print("If you want to check the trained result, please type commands below : ")
        print(
            "docker run -it --rm --gpus all -p 7007:7007 \
            -v /tmp/nerf_3d_reconstruction/"
            + self.result_id
            + ":/workspace dromni/nerfstudio:0.3.1"
            '/bin/bash -c "find -name config.yml | xargs -I {} ns-viewer --load-config {}"'
        )



from dataclass_wizard import JSONWizard
from typing import List, Annotated
from dataclasses import dataclass, field


@dataclass
class BoundingBox(JSONWizard):  # type: ignore
    x1: float = 0.0
    x2: float = 0.0
    y1: float = 0.0
    y2: float = 0.0


@dataclass
class BoundingBoxAnnotation(JSONWizard):  # type: ignore
    box: BoundingBox = BoundingBox()
    score: float = 0
    object_class: str = ""
    clip_embeddings: Annotated[List[float], 512] = field(default_factory=list)


@dataclass
class ImageAnnotation(JSONWizard):  # type: ignore
    image_index: int = 0
    bounding_boxes: List[BoundingBoxAnnotation] = field(default_factory=list)



from argparse import ArgumentParser
from amber_mcap.exception import TaskDescriptionError
import os
from yaml import safe_load  # type: ignore
from amber_mcap.automation.detic_image_labeler import DeticImageLabeler
from amber_mcap.automation.clip_image_annotation_filter import ClipImageAnnotationFilter
from amber_mcap.automation.nerf_3d_reconstruction import Nerf3DReconstruction
from amber_mcap.dataset.images_dataset import ImagesDataset, ReadImagesConfig
from amber_mcap.dataset.images_and_annotations_dataset import (
    ImagesAndAnnotationsDataset,
    ReadImagesAndAnnotationsConfig,
)
from amber_mcap.importer.video import VideoImporter, VideoImporterConfig
from amber_mcap.visualization.clip_embeddings_visualization import (
    ClipEmbeddingsVisualization,
)
from typing import Any, Callable, List
import torch


def setup_arguments_and_parser_for_automation(
    parser_automation: Any, func: Callable[[Any], None]
) -> None:
    parser_automation.add_argument(
        "task_description_yaml_path",
        help="Path to the yaml description file path",
        default="",
    )
    parser_automation.add_argument(
        "dataset_description_yaml_path",
        help="Path to the yaml description file path for dataset",
        default="",
    )
    parser_automation.add_argument(
        "rosbag_path",
        help="Path to the target rosbag path",
        default="",
    )
    parser_automation.add_argument(
        "output_rosbag_path",
        help="Path to the output rosbag path",
        default="",
    )
    parser_automation.set_defaults(handler=func)


def setup_arguments_and_parser_for_visualization(
    parser_visualization: Any, func: Callable[[Any], None]
) -> None:
    parser_visualization.add_argument(
        "task_description_yaml_path",
        help="Path to the yaml description file path",
        default="",
    )
    parser_visualization.add_argument(
        "dataset_description_yaml_path",
        help="Path to the yaml description file path for dataset",
        default="",
    )
    parser_visualization.add_argument(
        "rosbag_path",
        help="Path to the target rosbag path",
        default="",
    )
    parser_visualization.set_defaults(handler=func)


def check_config_files_exists_for_automation(args: Any) -> None:
    if not os.path.exists(args.task_description_yaml_path):
        raise TaskDescriptionError(
            "Task description yaml path : "
            + args.task_description_yaml_path
            + " does not exist."
        )
    if not os.path.exists(args.dataset_description_yaml_path):
        raise TaskDescriptionError(
            "Dataset description yaml path : "
            + args.dataset_description_yaml_path
            + " does not exist."
        )
    if not os.path.exists(args.rosbag_path):
        raise TaskDescriptionError(
            "Specified rosbag path : " + args.rosbag_path + " does not exist."
        )


def run_detic_image_labaler_automation(args: Any) -> None:
    check_config_files_exists_for_automation(args)
    labeler = DeticImageLabeler(args.task_description_yaml_path)
    dataset = ImagesDataset(
        args.rosbag_path,
        ReadImagesAndAnnotationsConfig.from_yaml_file(
            args.dataset_description_yaml_path
        ),
    )
    annotations = labeler.inference(dataset)
    labeler.write(
        dataset, "/detic_image_labeler/annotation", annotations, args.output_rosbag_path
    )


def run_clip_image_annotation_filter_automation(args: Any) -> None:
    check_config_files_exists_for_automation(args)
    filter = ClipImageAnnotationFilter(args.task_description_yaml_path)
    dataset = ImagesAndAnnotationsDataset(
        args.rosbag_path,
        ReadImagesAndAnnotationsConfig.from_yaml_file(
            args.dataset_description_yaml_path
        ),
    )
    annotations = filter.inference(dataset)
    filter.write(
        dataset,
        "/clip_image_annotation_filter/annotation",
        annotations,
        args.output_rosbag_path,
    )


def run_nerf_3d_reconstruction_automation(args: Any) -> None:
    check_config_files_exists_for_automation(args)
    task_description = {}
    with open(args.task_description_yaml_path, "rb") as file:
        task_description = safe_load(file)
    reconstruction = Nerf3DReconstruction(args.task_description_yaml_path)
    dataset = ImagesDataset(
        args.rosbag_path,
        ReadImagesConfig.from_yaml_file(args.dataset_description_yaml_path),
    )
    reconstruction.inference(dataset)


def setup_arguments_and_parser_for_import(
    parser_import: Any, func: Callable[[Any], None]
) -> None:
    parser_import.add_argument("video", help="Video file path for importer.")
    parser_import.add_argument("config", help="Config file for video importer.")
    parser_import.set_defaults(handler=func)


def check_config_files_exists_for_import(args: Any) -> None:
    if not os.path.exists(args.config):
        raise TaskDescriptionError(
            "Specified config : " + args.config + " does not exist."
        )


def run_video_import(args: Any) -> None:
    check_config_files_exists_for_import(args)
    VideoImporter(args.video, VideoImporterConfig.from_yaml_file(args.config)).write()


def check_config_files_exists_for_visualize(args: Any) -> None:
    if not os.path.exists(args.task_description_yaml_path):
        raise TaskDescriptionError(
            "Task description yaml path : "
            + args.task_description_yaml_path
            + " does not exist."
        )
    if not os.path.exists(args.dataset_description_yaml_path):
        raise TaskDescriptionError(
            "Specified config : " + args.config + " does not exist."
        )


def run_visualize_image(args: Any) -> None:
    check_config_files_exists_for_visualize(args)
    dataset = ImagesAndAnnotationsDataset(
        args.rosbag_path,
        ReadImagesAndAnnotationsConfig.from_yaml_file(
            args.dataset_description_yaml_path
        ),
    )
    visualization = ClipEmbeddingsVisualization(args.task_description_yaml_path)
    visualization.visualize(dataset)


def main() -> None:
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()
    # Setup command line options for automation
    parser_automation = subparsers.add_parser(
        "automation", help="Run automation command"
    )
    subparsers_automation = parser_automation.add_subparsers()
    # Setup command line option for detic image labaler
    parser_detic_image_labeler_automation = subparsers_automation.add_parser(
        "detic_image_labeler", help="Run detic image labaler for rosbag"
    )
    setup_arguments_and_parser_for_automation(
        parser_detic_image_labeler_automation, run_detic_image_labaler_automation
    )
    # Setup command line option for clip image annotation filter
    parser_clip_image_annotation_filter = subparsers_automation.add_parser(
        "clip_image_annotation_filter",
        help="Run clip image annotation filter for rosbag.",
    )
    setup_arguments_and_parser_for_automation(
        parser_clip_image_annotation_filter, run_clip_image_annotation_filter_automation
    )
    # Setup command line option for NeRF 3D reconstruction
    parser_nerf_3d_reconstruction = subparsers_automation.add_parser(
        "nerf", help="Run NeRF 3D reconstruction."
    )
    setup_arguments_and_parser_for_automation(
        parser_nerf_3d_reconstruction, run_nerf_3d_reconstruction_automation
    )

    # Setup command line options for visualization command
    parser_visualization = subparsers.add_parser(
        "visualize", help="Visualize rosbag data."
    )
    # Setup command line options for visualize image_embedding command
    subparsers_visualization = parser_visualization.add_subparsers()
    parser_visualization = subparsers_visualization.add_parser(
        "image_embedding", help="Visualize image embeddings in rosbag data."
    )
    setup_arguments_and_parser_for_visualization(
        parser_visualization, run_visualize_image
    )

    # Setup command line option for import command
    parser_import = subparsers.add_parser(
        "import", help="Import non-rosbag data into rosbag"
    )
    # Setup command line option for import video command
    subparsers_import = parser_import.add_subparsers()
    parser_video_import = subparsers_import.add_parser(
        "video", help="Import video images into rosbag"
    )
    setup_arguments_and_parser_for_import(parser_video_import, run_video_import)

    # Parsing arguments
    args = parser.parse_args()
    if hasattr(args, "handler"):
        args.handler(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from lavis.models.blip2_models.blip2_image_text_matching import Blip2ITM
from lavis.processors import load_processor
import torch
from pathlib import Path
from torchvision.io import read_image
from lavis.processors.blip_processors import BlipImageBaseProcessor
from torchvision import transforms
from PIL import Image
import torch.nn.functional as F
from lavis.models import load_model_and_preprocess
from typing import Optional


class Blip2ImageProcessor(BlipImageBaseProcessor):  # type: ignore
    def __init__(
        self,
        image_size: int = 384,
        mean: Optional[float] = None,
        std: Optional[float] = None,
    ) -> None:
        super().__init__(mean=mean, std=std)
        self.transform = transforms.Compose(
            [
                transforms.Resize(
                    (image_size, image_size),
                    interpolation=transforms.InterpolationMode.BICUBIC,
                ),
                self.normalize,
            ]
        )

    def __call__(self, item: torch.Tensor) -> torch.Tensor:
        return self.transform(item)


class Blip2Encoder:
    def __init__(self) -> None:
        self.device = torch.device("cuda") if torch.cuda.is_available() else "cpu"
        self.model, _, self.text_processors = load_model_and_preprocess(
            "blip2_image_text_matching", "pretrain", device=self.device, is_eval=True
        )
        self.image_processor = Blip2ImageProcessor(image_size=224)

    def encode_image(self, image: torch.Tensor) -> torch.Tensor:
        image = self.image_processor(image)
        assert image.dim() == 3
        image = image.unsqueeze(0).to(self.device)
        with self.model.maybe_autocast():
            image_embeds = self.model.ln_vision(self.model.visual_encoder(image))
        image_embeds = image_embeds.float()
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(
            image.device
        )
        query_tokens = self.model.query_tokens.expand(image_embeds.shape[0], -1, -1)
        query_output = self.model.Qformer.bert(
            query_embeds=query_tokens,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=image_atts,
            return_dict=True,
        )
        return F.normalize(
            self.model.vision_proj(query_output.last_hidden_state), dim=-1
        )

    def encode_text(self, text: str) -> torch.Tensor:
        text_tensor: torch.Tensor = self.model.tokenizer(
            text,
            truncation=True,
            max_length=self.model.max_txt_len,
            return_tensors="pt",
        ).to(self.device)
        text_output = self.model.Qformer.bert(
            text_tensor.input_ids,
            attention_mask=text_tensor.attention_mask,
            return_dict=True,
        )
        return F.normalize(
            self.model.text_proj(text_output.last_hidden_state[:, 0, :]), dim=-1
        )

    def encode_image_from_file(self, image_path: Path) -> torch.Tensor:
        return self.encode_image(transforms.ToTensor()(Image.open(image_path)))

    def get_cosine_similarity_from_image_and_text(
        self, image: torch.Tensor, text: str
    ) -> float:
        text_embedding = self.encode_text(text)
        image_embedding = self.encode_image(image)
        sim: float = 0.0
        sim, _ = torch.max(
            torch.bmm(image_embedding, text_embedding.unsqueeze(-1)), dim=1
        )
        return sim

    def get_cosine_similarity_from_image_file_and_text(
        self, image_path: Path, text: str
    ) -> float:
        return self.get_cosine_similarity_from_image_and_text(
            self.image_processor(transforms.ToTensor()(Image.open(image_path))), text
        )


if __name__ == "__main__":
    pass



from PIL import Image
from mcap.records import Message, Schema
from mcap_ros2.decoder import Decoder
from mcap_ros2._dynamic import DecodedMessage
from pyzstd import decompress, compress
from amber_mcap.exception import MessageDecodingError
import numpy as np
import torch
import torchvision.transforms as transforms
import numpy
from amber_mcap.unit.time import Time, TimeUnit
import math
from sys import byteorder
from typing import Any, List, Callable
import cv2
import io
import open3d
from struct import unpack
import tf2_amber


def compress_message(message: Message) -> Message:
    message.data = compress(message.data)
    return message


def decompress_message(message: Message) -> Message:
    message.data = decompress(message.data)
    return message


def ros_message_to_image(ros_message: DecodedMessage) -> Image:
    if "encoding" in dir(ros_message):
        match ros_message.encoding:
            case "rgb8":
                return Image.frombytes(
                    "RGB", (ros_message.width, ros_message.height), ros_message.data
                )
            case "8UC3":
                image = Image.frombytes(
                    "RGB", (ros_message.width, ros_message.height), ros_message.data
                )
                b, g, r = image.split()
                return Image.merge("RGB", (r, g, b))
            case _:
                raise MessageDecodingError(
                    "image_encodings in sensor_msgs/msg/Image is "
                    + ros_message.encoding
                    + " , it was not supported yet."
                )
    elif "format" in dir(ros_message):
        match ros_message.format:
            case "rgb8; jpeg compressed bgr8":
                jpeg_data = numpy.frombuffer(
                    ros_message.data,
                    dtype=np.uint8,
                    count=len(ros_message.data),
                    offset=0,
                )
                return Image.fromarray(
                    cv2.cvtColor(
                        cv2.imdecode(jpeg_data, flags=cv2.IMREAD_COLOR),
                        cv2.COLOR_BGR2RGB,
                    )
                )
            case _:
                raise MessageDecodingError(
                    "Unsupported compressed message format. Please check rosbag data."
                )
    else:
        raise MessageDecodingError(
            "The image message should have a sensor_msgs/msg/Image or sensor_msgs/msg/CompressedImage type."
        )


def get_pointcloud_field(field_name: str, fields: List[Any]) -> Any:
    for field in fields:
        if field.name == field_name:
            return field
    raise MessageDecodingError(
        "Fields : "
        + field_name
        + " does not existing in fields. Please check rosbag data."
    )


def read_float_value_from_pointcloud_data(
    field_name: str, fields: List[Any], data: np.array, is_bigendian: bool
) -> float:
    field = get_pointcloud_field(field_name, fields)

    def get_bytes(data: np.array, offset: int, size: int) -> bytes:
        return bytes(data[offset : offset + size].tobytes())

    # See also https://docs.ros2.org/foxy/api/sensor_msgs/msg/PointField.html
    # INT8
    if field.datatype == 1:
        if is_bigendian:
            return float(unpack(">b", get_bytes(data, field.offset, 1))[0])
        else:
            return float(unpack("<b", get_bytes(data, field.offset, 1))[0])
    # UINT8
    elif field.datatype == 2:
        if is_bigendian:
            return float(unpack(">B", get_bytes(data, field.offset, 1))[0])
        else:
            return float(unpack("<B", get_bytes(data, field.offset, 1))[0])
    # INT16
    elif field.datatype == 3:
        if is_bigendian:
            return float(unpack(">h", get_bytes(data, field.offset, 2))[0])
        else:
            return float(unpack("<h", get_bytes(data, field.offset, 2))[0])
    # UINT16
    elif field.datatype == 4:
        if is_bigendian:
            return float(unpack(">H", get_bytes(data, field.offset, 2))[0])
        else:
            return float(unpack("<H", get_bytes(data, field.offset, 2))[0])
    # INT32
    elif field.datatype == 5:
        if is_bigendian:
            return float(unpack(">i", get_bytes(data, field.offset, 4))[0])
        else:
            return float(unpack("<i", get_bytes(data, field.offset, 4))[0])
    # UINT32
    elif field.datatype == 6:
        if is_bigendian:
            return float(unpack(">I", get_bytes(data, field.offset, 4))[0])
        else:
            return float(unpack("<I", get_bytes(data, field.offset, 4))[0])
    # FLOAT32
    elif field.datatype == 7:
        if is_bigendian:
            return float(unpack(">f", get_bytes(data, field.offset, 4))[0])
        else:
            return float(unpack("<f", get_bytes(data, field.offset, 4))[0])
    # FLOAT64
    elif field.datatype == 8:
        if is_bigendian:
            return float(unpack(">d", get_bytes(data, field.offset, 8))[0])
        else:
            return float(unpack("<d", get_bytes(data, field.offset, 8))[0])
    else:
        raise MessageDecodingError(
            "Invalid data type : "
            + str(field.datatype)
            + " does not existing in fields. Please check rosbag data."
        )


# This function was implemented with reference to PointPillars
# (https://github.com/zhulf0804/PointPillars/blob/b9948e73505c8d6bfa631ffdf76c7148e82c5942/utils/io.py#L18-L24)
def ros_message_to_pointcloud(
    ros_message: DecodedMessage,
) -> torch.Tensor:
    pointcloud = open3d.geometry.PointCloud()
    points = numpy.frombuffer(
        ros_message.data, dtype=np.uint8, count=len(ros_message.data), offset=0
    ).reshape(
        [int(len(ros_message.data) / ros_message.point_step), ros_message.point_step]
    )
    tensor = torch.zeros((len(points), 4), dtype=torch.float32)
    for i, point in enumerate(points):
        tensor[i] = torch.Tensor(
            [
                read_float_value_from_pointcloud_data(
                    "x", ros_message.fields, point, ros_message.is_bigendian
                ),
                read_float_value_from_pointcloud_data(
                    "y", ros_message.fields, point, ros_message.is_bigendian
                ),
                read_float_value_from_pointcloud_data(
                    "z", ros_message.fields, point, ros_message.is_bigendian
                ),
                read_float_value_from_pointcloud_data(
                    "intensity", ros_message.fields, point, ros_message.is_bigendian
                ),
            ]
        )
    return tensor


def image_to_tensor(image: Image) -> torch.Tensor:
    return transforms.Compose([transforms.ToTensor()])(image)


def decode_message(message: Message, schema: Schema, decompress: bool) -> Message:
    decoder = Decoder()
    if decompress:
        return decoder.decode(schema, decompress_message(message))
    else:
        return decoder.decode(schema, message)


def decode_image_message(
    message: Message, schema: Schema, decompress: bool
) -> torch.Tensor:
    return image_to_tensor(
        ros_message_to_image(decode_message(message, schema, decompress))
    )


def decode_pointcloud_message(
    message: Message, schema: Schema, decompress: bool
) -> torch.Tensor:
    return ros_message_to_pointcloud(decode_message(message, schema, decompress))


def build_message_from_image(
    image: numpy.ndarray, frame_id: str, stamp: Time, image_encodings: str = "bgr8"
) -> Any:
    data = image.flatten()
    floored_stamp: float = math.floor(stamp.get(TimeUnit.SECOND))
    return {
        "header": {
            "stamp": {
                "sec": int(floored_stamp),
                "nanosec": Time(
                    stamp.get(TimeUnit.SECOND) - floored_stamp, TimeUnit.SECOND
                ).get(TimeUnit.NANOSECOND),
            },
            "frame_id": frame_id,
        },
        "height": int(image.shape[0]),
        "width": int(image.shape[1]),
        "encoding": image_encodings,
        "is_bigendian": True if byteorder == "big" else False,
        "step": len(data) // int(image.shape[0]),
        "data": data.tolist(),
    }


def build_transform_stamped_message(
    message: Message, schema: Schema, decompress: bool
) -> List[tf2_amber.TransformStamped]:
    tf_messages: Any = decode_message(message, schema, decompress)
    tf_amber_messages: List[tf2_amber.TransformStamped] = []
    for message in tf_messages.transforms:
        tf_amber_message = tf2_amber.TransformStamped(
            tf2_amber.Header(
                tf2_amber.Time(message.header.stamp.sec, message.header.stamp.nanosec),
                message.header.frame_id,
            ),
            message.child_frame_id,
            tf2_amber.Transform(
                tf2_amber.Vector3(
                    message.transform.translation.x,
                    message.transform.translation.y,
                    message.transform.translation.z,
                ),
                tf2_amber.Quaternion(
                    message.transform.rotation.x,
                    message.transform.rotation.y,
                    message.transform.rotation.z,
                    message.transform.rotation.w,
                ),
            ),
        )
        tf_amber_messages.append(tf_amber_message)
    return tf_amber_messages


def build_message_from_time(sec: int, nanosec: int):
    return {"sec": sec, "nanosec": nanosec}


def build_message_from_header(stamp: tf2_amber.Time, frame_id: str):
    return {
        "stamp": build_message_from_time(stamp.sec, stamp.nanosec),
        "frame_id": frame_id,
    }


def build_message_from_vector3(x: float, y: float, z: float):
    return {"x": x, "y": y, "z": z}


def build_message_from_quaternion(x: float, y: float, z: float, w: float):
    return {"x": x, "y": y, "z": z, "w": w}


def build_message_from_transform(
    translation: tf2_amber.Vector3, rotation: tf2_amber.Quaternion
):
    return {
        "translation": build_message_from_vector3(
            translation.x, translation.y, translation.z
        ),
        "rotation": build_message_from_quaternion(
            rotation.x, rotation.y, rotation.z, rotation.w
        ),
    }


def build_message_from_transform_stamped(
    header: tf2_amber.Header, child_frame_id: str, transform: tf2_amber.Transform
):
    return {
        "header": build_message_from_header(header.stamp, header.frame_id),
        "child_frame_id": child_frame_id,
        "transform": build_message_from_transform(
            transform.translation, transform.rotation
        ),
    }


def build_message_from_tf(transforms: List[tf2_amber.TransformStamped]):
    ret = {"transforms": []}
    for transform in transforms:
        ret["transforms"].append(
            build_message_from_transform_stamped(
                transform.header, transform.child_frame_id, transform.transform
            )
        )
    return ret
